/*
* Distributed Computational System
* Server Module
*
* $Id: server.c,v 1.7 2001/11/29 18:33:30 kris Exp $
*/

#include "server_plugin.h"
#include "protocol.h"
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <netinet/in.h>

extern int errno;

/*#define RUN_AS_DAEMON*/		/* run in the background */
#define EXIT_ON_COMPLETE		/* exit when there are no more chunks */
#define LOG_FILE "server.log"
#define PLUGIN_NAME "./sp_primes.so"
#define LISTEN_PORT 10001
#define MAX_CLIENTS 300

typedef struct _plugin_t {
	void *plug;
	sp_init_f sp_init;
	sp_getversion_f sp_getversion;
	sp_getproblemchunk_f sp_getproblemchunk;
	sp_completedchunk_f sp_completedchunk;
} plugin_t;

typedef struct _client_t {
	int sock;
	packet_buf_t buffer;
} client_t;

void fatal_error( const char *err );
void error( const char *err );
void log( const char *txt );
int load_plugin( const char *file, plugin_t *plugin );
int srv_setnonblocking( int sockfd );
int srv_listen( unsigned int port, int backlog, struct sockaddr_in *serv );
int srv_daemonize();
int handle_command( plugin_t *plug, command_t *command, client_t *client );


int main( int argc, char **argv )
{
	int lsock, maxfd, i, n;
	struct sockaddr_in serv, new_cli;
	plugin_t plug;
	fd_set rfds;
	client_t clients[MAX_CLIENTS];
	struct linger set_linger = { 1, 0 };

#ifdef RUN_AS_DAEMON
	FILE *log;
	srv_daemonize();
	log = fopen( LOG_FILE, "a" );
	stdout = stderr = log;
#endif

	if( load_plugin( PLUGIN_NAME, &plug ) < 0 )
		fatal_error( "Unable to load plugin" );
	(plug.sp_init)( 0 );
	for( i = 0; i < MAX_CLIENTS; i++ )
		clients[i].sock = -1;

	if( (lsock = srv_listen( LISTEN_PORT, 5, &serv )) < 0 )
		fatal_error( "Unable to listen" );

	setsockopt( lsock, SOL_SOCKET, SO_LINGER, (void *)&set_linger, sizeof( set_linger ) );

	while( 1 ) {
		maxfd = lsock;
		FD_ZERO( &rfds );
		FD_SET( lsock, &rfds );

		for( i = 0; i < MAX_CLIENTS; i++ ) {
			if( clients[i].sock != -1 ) {
				FD_SET( clients[i].sock, &rfds );
				if( clients[i].sock > maxfd )
					maxfd = clients[i].sock;
			}
		}

		n = select( maxfd + 1, &rfds, NULL, NULL, NULL );
		if( n < 0 ) {
			if( errno == EINTR )
				continue;
			fatal_error( "Select failed" );
		}
		if( FD_ISSET( lsock, &rfds ) ) {
			/* new client connection */
			int sa_len = sizeof( struct sockaddr_in );

			for( i = 0; i < MAX_CLIENTS && clients[i].sock != -1; i++ );

			if( i == MAX_CLIENTS ) {
				int tmp_sock;
				error( "Maximum number of clients reached" );
				tmp_sock = accept( lsock, (struct sockaddr *)&new_cli, &sa_len );
				close( tmp_sock );
			} else {
				clients[i].sock = accept( lsock, (struct sockaddr *)&new_cli, &sa_len );
				srv_setnonblocking( clients[i].sock );
				INIT_PACKET_BUF(clients[i].buffer);
			}
			n--;
		}
		for( i = 0; i < MAX_CLIENTS; i++ ) {
			if( n == 0 )
				break;
			if( clients[i].sock != -1 && FD_ISSET( clients[i].sock, &rfds ) ) {
				/* incoming data from client (or disconnection) */
				int kill = 0;
				int rv = PR_Recv( clients[i].sock, &clients[i].buffer );
				switch( rv ) {
				case -1:
					error( "PR_Recv error" );
printf("errno %d\n", errno);
				case 0:
					kill = 1;
					break;
				/* we can ignore case 1, it doesn't mean much */
				case 2:
				{
					command_t cmd = GETCMD(clients[i].buffer);
					if( handle_command( &plug, &cmd, &clients[i] ) < 0 )
						kill = 1;
					break;
				}
				}
				if( kill ) {
					close( clients[i].sock );
					clients[i].sock = -1;
					FREE_PACKET_DATA(clients[i].buffer);
					log( "Disconnected client" );
				}
				n--;
			}
		}
	}

	close( lsock );
	return 0;
}

void fatal_error( const char *err )
{
	fprintf( stderr, "FATAL ERROR: %s\n", err );
	exit( 1 );
}

void error( const char *err )
{
	fprintf( stderr, "ERROR: %s\n", err );
}

void log( const char *txt )
{
	fprintf( stderr, "%s\n", txt );
}

int load_plugin( const char *file, plugin_t *plugin )
{
	char *error;

	plugin->plug = dlopen( file, RTLD_LAZY );
	if( !plugin->plug )
		return -1;
	dlerror(); /* clear error */
	plugin->sp_init = dlsym( plugin->plug, "SP_Init" );
	if( (error = dlerror()) != NULL )
		return -1;
	plugin->sp_getversion = dlsym( plugin->plug, "SP_GetVersion" );
	if( (error = dlerror()) != NULL )
		return -1;
	plugin->sp_getproblemchunk = dlsym( plugin->plug, "SP_GetProblemChunk" );
	if( (error = dlerror()) != NULL )
		return -1;
	plugin->sp_completedchunk = dlsym( plugin->plug, "SP_CompletedChunk" );
	if( (error = dlerror()) != NULL )
		return -1;
	log( "Successfully loaded plugin" );
	return 0;
}

int srv_setnonblocking( int sockfd )
{
	long v;

	v = fcntl( sockfd, F_GETFL, 0 );
	if( v & O_NONBLOCK )
		return 0;
	return fcntl( sockfd, F_SETFL, v | O_NONBLOCK );
}

int srv_listen( unsigned int port, int backlog, struct sockaddr_in *serv )
{
	int fd = -1;
	int i;

	fd = socket( AF_INET, SOCK_STREAM, 0 );
	if( fd < 0 )
		return -1;
	i = 1;
	if( setsockopt( fd, SOL_SOCKET, SO_REUSEADDR, (char *)&i, sizeof( &i ) ) < 0 )
		goto srv_listen_err;
	memset( serv, 0, sizeof( struct sockaddr_in ) );
	serv->sin_family = AF_INET;
	serv->sin_addr.s_addr = htonl( INADDR_ANY );
	serv->sin_port = htons( (unsigned short)port );

	if( bind( fd, (struct sockaddr *)serv, sizeof( struct sockaddr_in ) ) < 0 )
		goto srv_listen_err;
	if( listen( fd, backlog ) < 0 )
		goto srv_listen_err;
	if( srv_setnonblocking( fd ) < 0 )
		goto srv_listen_err;

	return fd;
srv_listen_err:
	close( fd );
	return -1;
}

int srv_daemonize()
{
	int i;

	for( i = 0; i < NOFILE; i++ )
		close( i );
	if( fork() > 0 )
		exit( 0 );
	setpgid( 0, getpid() );
        return 0;
}

int handle_command( plugin_t *plug, command_t *command, client_t *client )
{
	server_chunk_t chunk;
	command_t cmd;

	memset( &cmd, 0, sizeof( command_t ) );
	memset( &chunk, 0, sizeof( server_chunk_t ) );

	switch( command->cmd ) {
	case 12:
		/* request assignment (no arguments) */
		if( (plug->sp_getproblemchunk)( &chunk ) < 0 )
#ifdef EXIT_ON_COMPLETE
			exit( 0 );
#else
			return -1;
#endif
		cmd.cmd = 10; /* assign problem */
		cmd.len = 24 + chunk.len;
		cmd.data = (unsigned char *)malloc( cmd.len );
		memcpy( cmd.data, &chunk, 24 );
		memcpy( cmd.data + 24, chunk.data, chunk.len );
		FREE_CHUNK_DATA(chunk);
		if( PR_Send( client->sock, 0, &cmd ) < 0 )
			error( "Unable to send response to assignment request" );
		FREE_COMMAND_DATA(cmd);
		return 0;
	case 20:
		/* report results [chunk-id][chunk-type][data-len][data] */
		if( command->len < 21 )
			return -1;
		memcpy( chunk.chunk_id, command->data, 16 );
		chunk.chunk_type = command->data[16];
		chunk.problem_id = 0;
		memcpy( &chunk.len, command->data + 17, 4 );
		chunk.data = command->data + 21;
		if( (plug->sp_completedchunk)( &chunk ) < 0 )
			return -1;
		if( PR_Send( client->sock, 0, &MAKECMD(20,0,"") ) < 0 )
			error( "Unable to send response to result report" );
		return 0;
	default:
		return -1;
	}
}
